'''
Синхронизатор состояния оборудования и показаний датчиков варианта LITE и
БД Redis с минимальным UI на основе curses
v.0.3.0.2
-------------------------------------------------------------------------------
История изменений и "напоминашки" - в конце файла
OldBean, 14.12.2019
'''
#-------------------------------------------------------------------------------
import time, sys, os
import traceback
from threading import Thread
import io     # used to create file streams
import fcntl  # used to access I2C parameters like addresses
import smbus # С датчиками давления общаемся при помощи этой библиотеки
import math
import inspect
import redis
import re
import RPi.GPIO as gpio
import marshal
import curses
import builtins
builtins.prnFlag = False # Отменяем печать словаря созданных в lite объектов
from lite import *
#-------------------------------------------------------------------------------
RESET_PIN = 11 # Пин малинки для сброса модулей по линии Res
INT_PIN = 29   # Пин малинки для приема сигнала аварии по линии Int
#-------------------------------------------------------------------------------
# Вспомогательные функции
#-------------------------------------------------------------------------------
def rpushlpop(key, val, flag):
  '''
  rpushlpop(key, val, flag)
  Реализация кольцевых списков в БД redis.
  key - ключ БД, val - вставляемое значение, flag - индикатор типа списка
  (True - кольцевой, False - обычный)
  '''
  rdb.rpush(key, val) # Вставляем в конец списка
  if flag: # А для кольцевого списка удаляем первый элемент
    rdb.lpop(key)
'''--------------------------------------------------------------------------'''
status = '' # Строка статуса
def signalInt(channel): # Обработка прерывания по сигналу аварии Int
  global status
  ''' По сигналу INT обнуления контроллеров происходит аппаратно.
  Для согласования, обнулим текущие значения и в БД и перейдем в режим
  мониторинга'''
  for un in su.keys(): # Выключаем все контроллеры
    if un[0] in ['r', 'w', 'q']:
      su[un].val = 0.0
  rdb.set('mode', 0) # Перейдем в режим мониторинга (0)
  rdb.set('init', 1) # Выставляем флаг инициализации нового режима
  status = tform(time.time() - tst) + ' Получен сигнал INT'
  try: stdscr.addstr(23, 0, status, curses.color_pair(1))
  except: print('\033[91m' + status +'\033[0m')
'''--------------------------------------------------------------------------'''
def getma(key, ma, n):
  '''Вовращает текущее значение скользящего среднего
  key - ключ серии в БД
  ma  - предыдущее значение скользящего среднего
  n   - окно (кол-во отсчетов в среднем'''
  try: f = float(rdb.lindex(key, -1)) # Текущее значение
  except: f = 0.0
  try: fmn = float(rdb.lindex(key, -n - 1)) # Первое значение предыд. среднего
  except: fmn = 0.0
  ni = 1.0/n
  return ma - fmn*ni + f*ni
'''--------------------------------------------------------------------------'''
gpio.setmode(gpio.BOARD) # Нумерация пинов - как на плате
gpio.setup(INT_PIN, gpio.IN) # Порт для приема по линии Int
gpio.add_event_detect(INT_PIN, gpio.FALLING, callback = signalInt)
#gpio.setup(RESET_PIN, gpio.OUT) # Порт для сброса модулей по линии Res
# Реализовать Reset после согласования уровней 5 и 3.3 В !!!!!!!!!!!!!!!!
'''--------------------------------------------------------------------------'''
import argparse
parser = argparse.ArgumentParser()
parser.add_argument('-r', '--reset', action = 'store_true',
                    help = 'Сброс БД и имен устройств при инициализации')
parser.add_argument('-i', '--ip', default = 'localhost',
                    help = 'IP-адрес хоста с БД redis')
parser.add_argument('-p', '--port', default = 6379,
                    help = 'Порт хоста с БД redis')
parser.add_argument('-d', '--db', default = 0,
                    help = 'Номер рабочей базы данных')
'''Парсим командную строку и соединяемся с БД'''
rflag = parser.parse_args().reset # Флаг очистки БД и переименования устройств
host = parser.parse_args().ip # IP-адрес хоста с редиской
port = parser.parse_args().port # Порт хоста с редиской
db = parser.parse_args().db # Номер рабочей базы данных
rdb = redis.StrictRedis(host, port, db) # Соединяемся с БД
dbLen = 172800 # Размер циклической БД (по умолчанию 4-суточная история)
rrdFlag = False # Циклическая БД (Round-robin Database - RRD)
'''Чистим БД от предудущих сессий. Если флаг rflag выставлен очищается вся БД,
если нет, то сохраняется поле с исходником пользовательским скрипта
(\'uscript\') и хэш с конфигурацией (\'cfg\')'''
for key in rdb.keys():
  if (key in [b'cfg', b'uscript']) and (not rflag):
    continue
  rdb.delete(key)
for key in codes.keys(): # Заносим в БД коды типов устройств
  rdb.hset(b'codes', int(key), codes[key][0])
for unit in units: # и список единиц измерения
  rdb.rpush(b'units', unit)
if rdb.exists('uscript'): rdb.set('usinit', 1)
else: rdb.set('usinit', 0)

'''di - вспомогательный словарь для индексирования имен. Ключи представляют
собой первые буквы имен устройств, а значения - максимальное значение индекса
имен соотвествующих устройств'''
di = {}
for key in codes.keys():
  di[codes[key][1]] = -1
'''Найдем максимальные значения индексов в именах устройств на текущий
момент. Данные берутся из хэша \'cfg\' в БД. Ключи у этого хэша - hard-ключи
устройств, использовавшихся в предыдущих сессиях'''
for hkey in rdb.hkeys('cfg'):
  name = rdb.hget('cfg', hkey).decode()
  i = int(name[1:])
  if i > di[name[0]]:
    di[name[0]] = i
'''Теперь присвоим имена физическим величинам (P0, T0, T1, w0 и т.п.),
соответствующим найденным устройствам. Если устройство уже использвалось
в предыдущих сессиях и осталось на старом месте (т.е. остался тот же самый
hard-ключи (адрес,  канал и тип)), то ему присваивается его прежнее имя из
БД (хэш \'cfg\'), а если нет - формируется новое имя.
Заодно в этом же цикле формируем словарь обнаруженных устройств (su)
с пользовательскими именами в качестве ключей. Значениями в su являются
соответствующие объекты. На всякий случай заведем еще один вспомогательный
словарик unhk {пользовательское имя: hard ключ}'''
su = {}; unhk = {}

for hkey in sorted(obj.keys()):
  if rdb.hexists('cfg', hkey): # Устр-во на старом месте - берем старое имя
    obj[hkey][1] = rdb.hget('cfg', hkey).decode()
  else: # Новое (или замененное на другой тип) устройство
    let = codes[obj[hkey][0].fc][1] # Первая буква имени устр-ва данного типа
    di[let] += 1 # Модифицируем индекс
    name = let + ('%d'%(di[let])) # Формируем имя устройства
    obj[hkey][1] = name # Именуем
    rdb.hset('cfg', hkey, name) # и фиксируем новое устройства в БД
    print('Новое устройство: ',
          hkey, codes[obj[hkey][0].fc][0], ' Нарекаем: ', obj[hkey][1])
  su[obj[hkey][1]] = obj[hkey][0]
  unhk[obj[hkey][1]] = hkey
'''Формируем хэш калибровок в БД. Калибровки в БД хранятся в упакованном виде
в таком же формате, как и в EEPROM устройств'''
kl = rdb.hkeys('calibr')
if len(kl) > 0:
  rdb.hdel('calibr', *kl) # Очищаем хэш
rdb.hset('calibr', 'time', ser(1.0, 0, 7)) # Время
for un in sorted(su.keys()): # Остальные устройства
  if un[0] == 'H': # Пропускаем хабы
    continue
  buf = ser(*(su[un].calibr)) # Сериализация калибровочных данных устройства
  rdb.hset('calibr', un, buf) # Пишем в хэш калибровок

'''Генерирование класса режима работы Mode'''
codeMode = '\nclass Mode(object):\n  fc = 0xFC\n'
codeMode += '  def __init__(self, name'
'''Сформируем список имен исполнительных устройств'''
unl = []
for un in rdb.hkeys('calibr'):
  und = un.decode()
  if und[0] in ['w', 'q']:
    unl.append(und)
for un in unl: # Список параметров
  codeMode += ', ' + un + ' = 0.0'
codeMode += ', ss = 0):\n'
codeMode += '    self.name = ' + 'name\n'
for un in unl:
  codeMode += '    self.' + un + ' = ' + un + '\n'
codeMode += '    self.ss = ss\n'
codeMode += '  def initialize(self):\n'
for un in unl: # Операторы инициализации режима
  codeMode += '    rdb.rpush(\'must\', \'' + un + '\')\n'
  codeMode += '    rdb.rpush(\'must\', self.' + un + ')\n'
codeMode += '    rdb.set(\'ss\', self.ss)\n'
#print(codeMode)
exec(codeMode) # Помещаем класс Mode в пространство имен синхронизатора
#for nam in inspect.getmembers(sys.modules[__name__], inspect.isclass):
#  print(nam[0])

'''Далее будем работать с консолью синхронизатора в оконном режиме (curses).
Объявляем функции вывода информации в главном окне'''
'''--------------------------------------------------------------------------'''
def showHard():
  '''Выводим информацию о пользовательских именах, hard-ключах и калибровках'''
  stdscr.erase()
  col = 2; raw = 0
  stdscr.addstr(raw, col - 1, 'имя    название устройства')
  mxl = max([len(c[0]) for c in codes.values()]) + 6
  stdscr.addstr(raw, col + mxl - 1, 'hard-код   калибровка')
  raw += 1
  for un in sorted(su.keys()):
    if un[0] == 'H': # Пропускаем хабы
      continue
    stdscr.addstr(raw, col, un + '  ' + codes[su[un].fc][0])
    stdscr.addstr(raw, col + mxl, unhk[un] + '  ' +
                  str(deser(rdb.hget('calibr', un))))
    raw += 1
#  raw += 1
  for un in sorted(su.keys()):
    if un[0] == 'H': # Выводим информацию о хабах
      stdscr.addstr(raw, col, un + '  ' + codes[su[un].fc][0])
      stdscr.addstr(raw, col + mxl, unhk[un])
      raw += 1

  stdscr.addstr(22, 1, 'c - закрыть это окно')
  stdscr.refresh()

  global shFlag
  shFlag = True
'''--------------------------------------------------------------------------'''
def tform(t):
  '''Форматирование времени'''
  s = t % 60; t = t // 60
  m = t % 60; t = t // 60
  h = t % 24; d = t // 24
  return '%01d:%02d:%02d:%02d'%(d, h, m, s)
'''--------------------------------------------------------------------------'''
def showInfo():
  '''Окно для вывода текущей информации и управления процессом. Этот упрощенный
  вариант управления позволяет обойтись без специального клиентского приложения.
  В этом случае пользовательский скрипт может быть загружен непосредственно
  синхронизатором, так и сторонним загрузчиком'''
  stdscr.erase()

  raw = 0; col = 1
  stdscr.addstr(raw, col, 'Датчики:'); raw += 1
  for un in sorted(su.keys()):
    if un[0] in ['H', 'Z', 'q', 'w']: continue
    stdscr.addstr(raw, col, un)
    if un[0] in ['P', 'U']: form = '% 7.1f %s'
    elif un[0] in ['T']: form = '% 7.3f %s'
    else: form = '% 7.0f %s'
    try: val = float(rdb.lindex(un, -1))
    except: val = 0
    try: uni = units[int(deser(rdb.hget('calibr', un))[2])]
    except: uni = ''
    stdscr.addstr(raw, col + 5, form%(val, uni)); raw += 1

  raw += 1; uniq = ''
  stdscr.addstr(raw, col, 'Контроллеры:'); raw += 1
  for un in sorted(su.keys()):
    if not un[0] in ['q', 'w']: continue # Хабы и пустые каналы пропускаем
    stdscr.addstr(raw, col, un)
    form = '% 7.0f %s'
    try: val = float(rdb.lindex(un, -1))
    except: val = 0
    try: uni = units[int(deser(rdb.hget('calibr', un))[2])]
    except: uni = ''
    stdscr.addstr(raw, col + 5, form%(val, uni)); raw += 1
    if un[0] == 'q':
      try: uniq = units[int(deser(rdb.hget('calibr', un))[2])]
      except: uniq = ''
  stdscr.addstr(raw, col, '<q>')
  stdscr.addstr(raw, col + 5, '% 7.0f %s'%(ma, uniq)); raw += 1

  raw = 0; col = 26
  stdscr.addstr(raw, col, 'Режимы:'); raw += 1
  for i, mo in enumerate(modes):
    m = mo.name
    if i == int(rdb.get('mode')):
      stdscr.addstr(raw, col, str(i), curses.color_pair(2))
      stdscr.addstr('  ' + m, curses.color_pair(2))
    else:
      stdscr.addstr(raw, col, str(i))
      stdscr.addstr('  ' + m)
    raw += 1
  '''Пробуем вызвать продолжение интерфейса в пользовательском скрипте'''
  try: raw = ui(raw, col) # Колонка элементов управления
  except: pass

  raw += 2

  raw = 0; col = 46
  stdscr.addstr(raw, col, 'Время:'); raw += 1
  try: tim = float(rdb.lindex('time', -1)) # Общее время работы lsync
  except: tim = 0
  try: mtim = float(rdb.lindex('mtime', -1)) # Длительность текущего режима
  except: mtim = 0
  stdscr.addstr(raw, col, 'Общее  ' + tform(int(tim))); raw += 1
  stdscr.addstr(raw, col, 'Режима ' + tform(int(mtim))); raw += 2

  stdscr.addstr(raw, col, 'Управление:'); raw += 1

  if len(modes) > 1:
    stdscr.addstr(raw, col, '0,1,2... - выбор режима'); raw += 1
  '''Пытаемся дополнить справку пользовательскими командами'''
  try: raw = uihelp(raw, col)
  except: pass
  stdscr.addstr(raw, col, 'h - информация об оборудовании'); raw += 1
  stdscr.addstr(raw, col, 'u - загрузить uscr.py'); raw += 1
#  stdscr.addstr(raw, col, 'r - аппаратный сброс всех модулей'); raw += 1
  stdscr.addstr(raw, col, 'q - выход'); raw += 1
#  stdscr.addstr(23, 0, '%d %d'%(_x, _y))
  stdscr.addstr(23, 0, status, curses.color_pair(1))
  stdscr.refresh()
'''--------------------------------------------------------------------------'''
def loadUSF(fn):
  '''Загрузка пользовательского скрипта в поле \'uscript\' и установка флага
  преобразования и инициализации \'usinit\''''
  with open(fn) as f:
    code = f.read() # Загружаем файл со скриптом в строку code
  try: # Проверим пользовательский скрипт на ошибки Python
    compile(code, '', 'exec')
  except BaseException as e:
    with open('err', 'w') as f:
      f.write('В скрипте есть ошибки:')
      f.write('\n' + str(e))
      f.write('\nЗагрузка пользовательского скрипта отменена\n')
    return
  rdb.set('uscript', code) # Помещаем пользовательский скрипт в БД
  rdb.set('usinit', 1) # Выставляем флаг инициализации скрипта
  global modes
  del modes[1:] # Очищаем список режимов (все, кроме мониторинга)
  with open('log', 'a') as f:
    f.write('Скрипт загружен в память...\n')
'''--------------------------------------------------------------------------'''
def prepareUS():
  code = rdb.get('uscript').decode()
  '''Уберем все комментарии с ковычками'''
  pure = re.sub('"""', "'''", code)
  while pure.find("'''") >= 0:
    pure = re.sub("'''.*?'''", '', pure, re.DOTALL, re.S)
  '''Уберем все однострочные комментарии (с решеткой)'''
  slst = []
  for s in pure.split('\n'):
    scut = re.sub('#.*', '', s)
    if scut.strip(' \t') == '': continue
    slst.append(scut)
  pure = '\n'.join(slst)

  with open('log', 'a') as f:
    f.write(pure)

  global usbcode
  usbcode = compile(pure, '', 'exec')
'''--------------------------------------------------------------------------'''
def mui():
  '''Функция обеспечивает интерфейс с пользователем в консоли, в которой
  запущен синхронизатор. Работает в другом потоке'''
  global exFlag, shFlag
  while True:
    ch = stdscr.getch()
    if ch == ord('q'): # Завершение работы
      exFlag = True
      break
    if ch == ord('r'): # Аппаратный сброс модулей (по линии Res)
      RESET_PIN - 0
      time.sleep(0.1)
      RESET_PIN - 1
      break
    elif ch == ord('h'): # Показать информацию об оборудовании
      showHard()
    elif ch == ord('c'): # Закрыть окно с информацией об оборудовании
      shFlag = False
    elif ch == ord('u'): # Загрузить пользовательский скрипт
      loadUSF('uscr.py')
      prepareUS()
    elif ch >= 0x30 and ch < 0x30 + len(modes):
      rdb.set('mode', ch - 0x30)
    else:
      try: uicallback(ch) # Пытаемся вызвать пользовательскую callback-функцию
      except: pass

    if not shFlag:
      showInfo()
'''--------------------------------------------------------------------------'''
'''Удаляем старый файл ошибок'''
try:
  with open('err', 'r') as f:
    pass
  os.remove('err')
except: pass
'''Очищаем старый файл логов'''
with open('log', 'w') as f:
  pass

modes = []
modes.append(Mode('Мониторинг'))
mode = 0
rdb.set('mode', mode)
if int(rdb.get('usinit')) != 0: prepareUS()
else: usbcode = compile('pass', '', 'exec')

'''Начинаем работать в оконном режиме'''
try:
  '''Инициируем curses и консоль'''
  stdscr = curses.initscr() # Основное окно curses
  curses.start_color() # Будем использовать цвет для выделения
  curses.curs_set(0) # Убираем изображение курсора
  curses.noecho() # Убираем эхо при нажатии клавиш
  curses.cbreak() # Убираем необходимость Enter для ввода с клавиатуры
  stdscr.keypad(True) # Возможность ввода спецсимволов (PgUp, Left etc)
  stdscr.nodelay(True) # Неблокирующий режим (не ждем getch())

  curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
  curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
  curses.init_pair(3, curses.COLOR_BLUE, curses.COLOR_BLACK)
  curses.init_pair(4, curses.COLOR_CYAN, curses.COLOR_BLACK)
  curses.init_pair(5, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
  curses.init_pair(6, curses.COLOR_YELLOW, curses.COLOR_BLACK)

  '''Система всегда начинает работу с режима 0 (мониторинг). Текущее состояние
  системы хранится в поле \'mode\'. Режим может быть изменен внешней процедурой
  (непосредственно в БД) или пользовательским скриптом, при выполнении условий,
  определенных логикой процесса. История режимов сохраняется в списке с ключем
  \'moden\'. Если mode != moden[-1]  - запускается процедура обработки смены
  режима работы'''
  rdb.set('mode', 0) # Начальное состояние системы - всегда 0 (мониторинг)

  dt = 2 # Длительность такта в сек
  cntr = 0; # Счетчик тактов синхронизации
  tst = time.time() # Начальный момент времени
  mtst = tst # Момент начала текущего режима
  rdb.set('tst', tst) # Запомним его в БД на всякий случай
  rdb.set('status', 'r') # Статус lsync: r - работа, q - выход
  codep = None # Байт-код пользовательского скрипта
  rdb.set('exep', 0) # Флаг разрешения исполнения пользовательского срипта
  shFlag = False # Флаг показа информации об оборудовании
  exFlag = False # Флаг выхода из главного цикла
  ma = 0 # Скользящее среднее значение скорости отбора
  nma = 300 # Ширина окна скользящего среднего (10 мин)

  thread1 = Thread(target = mui) # Запускаем UI в новом потоке
  thread1.start()
  while True: # Глобальный цикл синхронизации БД и оборудования
    now = time.time(); end = now + dt
    rdb.set('lock', 1) # Флаг "мягкой" блокировки внешнего доступа к БД
    '''1. Первый шаг в глобальном цикле - запуск преобразований датчиков.
    Эта команда должна быть дана: хабам, которые запускают циклы преобразований
    подключенных к ним датчиков, датчикам RMS, атмосферного давления и другим
    объектам, у которых реализован метод conv()'''
    for un in su.keys(): # Запуск преобразователей
      try:
        su[un].conv()
      except AttributeError: pass # Игнорируем это исключение
      except BaseException as error: print('\n', str(error))
    '''2. Пока (на уровне оборудования) выполняются преобразования, просмотрим
    устройства, требующие обновления: это контроллеры и интегрирующие датчики,
    имеющие опцию "сброс" (как, например, датчик расхода, ассоциированный с
    контроллером PWM. Пользовательские имена устройств, требующих обновления,
    и сами новые значения, которые будут присвоены контроллерам, находятся в
    очереди в БД с ключом \'must\'. Пары \'имя - новое значение\' помещаютмя в
    очередь справа. Сначала имя, затем - значение. Извлекаются слева,
    естественно, в том же порядке: имя потом значение'''
    while rdb.llen('must') > 0:
      un = rdb.lpop('must').decode()
      val = float(rdb.lpop('must'))
      su[un].val = val
    '''3. Для актуализации физических величин в некоторых объектах (например,
    хабах) необходимо вызвать метод update(). Этот метод может ожидать
    окончания преобразования. Поэтому возможны задержки (до скунды). В это
    время lsync не работает с БД. Поэтому "мягкую" блокировку БД, перед вызовом
    update(), можно снять, разрешая работу с БД внешним приложениям'''
    rdb.set('lock', 0)
    for un in su.keys():
      try:
        su[un].update()
      except AttributeError: pass # Игнорируем это исключение
      except BaseException as error: print('\n', str(error))
    '''4. Теперь можно выполнить синхронизацию БД с состоянием контроллеров и
    показаниями датчиков'''
    rdb.set('lock', 1) # Опять выставим флаг "мягкой" блокировки БД
    for un in sorted(su.keys()):
      if un[0] == 'H': continue # Игнорируем хабы
      rpushlpop(un, su[un].val, rrdFlag) # Вставим значение в соотв. списки БД
    '''5. Последний этап синхронизации \'железа\' с БД - обновление калибровок.
    Список пользовательских имен устройств, для которых требуется обновление
    калибровок находится в БД с ключом \'cmust\'. Новые значения находятся
    в хэше \'calibr\' с соответствующими ключами. Изменение калибровочных
    данных допускается только в режиме мониторинга (режим 0). Если запрос
    (ненулевой список \'cmust\') поступил в другом режиме - он игнорируется,
    а измененные в хэше \'calibr\' калибровочные данные восстанавливаются.'''
    if rdb.llen('cmust') > 0: # Запрос на изменения калибровочных данных
      if int(rdb.get('mode')) != 0: # Отклоняем запрос
        print('Изменение калибровочных данных игнорируется, т.к. mode != 0')
        while rdb.llen('cmust') > 0: # Восстановим калибровочные данные в БД
          un = rdb.lpop('cmust').decode()
          buf = ser(*(su[un].calibr)) # Сериализация старых калибр. данных
          rdb.hset('calibr', un, buf) # Восстановим значения в хэше калибровок
      else: # Режим мониторинга - можно обновить калибровки в устройствах
        while rdb.llen('cmust') > 0:
          un = rdb.lpop('cmust').decode()
          calw = deser(rdb.hget('calibr', un)) # Список новых калибр. данных
#          su[un].wc(*calw)
          su[un].calibr = calw
    rdb.set('lock', 0) # Теперь можно окончательно снять "мягкую" блокировку БД
    '''6. На этом этапе цикла у синхронизатора есть возможность выполнить
    пользовательский скрипт, в котором собственно и должна быть реализована
    логика автоматического управления процессом. Пользовательский скрипт должен
    быть написан на языке python. Выражения могут содержать:
    1) все переменные, хранимые синхронизатором в БД, имеющие пользовательские
    имена (физические величины, измеряемые датчиками, присваиваемые
    контроллерам), время (общее и время работы текущего режима) и т.п.;
    2) пользовательские параметры (константы и переменные), определяемые в
    тексте пользовательского скрипта;
    3) пользовательские функции, которые можно использовать в пользовательских
    скриптах;
    4) блок пользовательских условий для перехода к другим режимам работы
    установки или аварийном завершении.
    Загрузка пользовательского скрипта может быть выполнена в любой момент,
    при условии, что система находится в режиме мониторинга (mode = 0).
    При этом загруженные ранее скрипты и пользовательские данные удаляются
    из БД.'''

    exec(usbcode) # Выполняем пользовательский скрипт
    ma = getma('q0', ma, nma) # Вычисляем скользящее целое

    '''7.На завершающем этапе такта производится обработка смены режима работы,
    запись в БД текущего времени, длительности текущего режима и номера режима.
    Эта же информация выводится и на консоль.'''
    mode = int(rdb.get('mode'))
    if(rdb.llen('moden') > 0):
      if mode != int(rdb.lindex('moden', -1)): # Произошла смена режима
        modes[mode].initialize() # Инициализация нового режима
        for un in su.keys(): # Сброс интеграторов (пока только расход)
          if un[0] in ['Q']:
            su[un].val = 0.0
        mtst = now

    rpushlpop('time', now - tst, rrdFlag) # Пишем в БД относительное время
    rpushlpop('mtime', now - mtst, rrdFlag) # Длительность текущего режима
    rpushlpop('moden', mode, rrdFlag) # Номер текущего режима

    while end - time.time() > 0: # Ждем окончания интервала
      if exFlag: break
    if exFlag: break
except Exception as e:
  traceback.print_exc()
  with open('err', 'w') as f:
    f.write(str(e))
finally:
  thread1.join()
  stdscr.keypad(False)
  curses.nocbreak()
  curses.echo()
  curses.endwin()
  for un in su.keys(): # Выключаем контроллеры при завершении работы lsync
    if un[0] in ['r', 'w', 'q']:
      su[un].val = 0.0
  gpio.cleanup() # Очистка установок gpio
#-------------------------------------------------------------------------------
'''
02.08.2019 Ревизия старой версия (0.1.1.0) синхронизатора. Переписаны классы-
           обертки устройств на шине i2c с учетом новых прошивок (на время
           критических операций модуля интерфейс i2c с ним недоступен).
07.08.2019 Изменена политика именования ключей устройств в Redis, упрощен
           цикл синхронизации. Опция калибровки через Redis удалена.
09.08.2019 Написаны и отлажены классы BMP180 и BMP280. Модуль Adafruit_BMP
           больше не используется.
11.08.2019 Написан и отлажен класс MS5611 для соответствующего датчика
           атмосферного давления.
18.08.2019 Добавлена обработка сигнала аварии по линии Int. Коррекция только
           в БД, т.к. "железо" выключается (обнуляется) аппаратно.
24.08.2019 Синхронизатор существенно переработан. Новая версия - 0.3.0.0.
           Для всех датчиков теперь есть соответствующие классы. В том числе
           и класс виртуального датчика расхода Vol. В результате исчезли
           многие ветвления, вызванные неоднородностью классов устройств.

           Изменена логика присваивания пользовательских имен для устройств
           (физических величин). Теперь имена менее чувствительны к изменениях
           конфигурации "железа"

           Логика присваивания значений тоже изменена. Поле с ключом 'must'
           теперь формирует очередь устройств, состояние которых нужно обновить
           (контроллеры, интегральные датчики со сбросом). В очередь справа
           помещаются пары \'имя - значение\' (сначала имя, затем значение).
           Извлекаются эти данные из списка \'must\' слева.

           Для всех датчиков (как реальных, так и виртуальных) теперь есть свои
           собственные классы с геттерами и сеттерами для интегрирующих
           датчиков.
01.09.2019 Добавлен и "вчерне" отлажен функционал, связанный с выполнением
           пользовательских скриптов
14.09.2019 Много мелких изменений, связанных с отладкой функционала, связанного
           с исполнением поьзовательских скриптов
10.10.2019 Объявления классов вариант LITE выделены в отдельный модуль lite.py
19.10.2019 Реализация и черновая отладка UI на базе curses для локальной работы
           с синхронизатором. UI в сочетании с соответствующими функциями в
           пользовательском скрипте позволяет проводить ректификацию не
           используя специальных клиентских приложений.
01.11.2019 Финальное тестирование (вторая ректификация), мелкие правки.
           Добавление вычисления и индикации скользящего среднего для скорости
           отбора для контроля скважности старт-стопа
14.11.2019 Мелкие правки, устранение некритических ошибок при загрузке
           пользовательских скриптов, вызванные расширением их функцинала
14.12.2019 Добавлена обработка исключения (index error), возникающего
           при отсутствии калибровочных данных
'''
#-------------------------------------------------------------------------------
''' Планы:
    1. Согласовать уровень пина Res малинки (не более 3.3В) и линии Res общей
       шины (5В) и реализовать аппаратный сброс всех модулей по команде
       малинки (клавиша "r" UI)
    2. Все выдачи об ошибках, сбоях оборудования и т.п. направить в файл log
    3. Реализовать вывод ближайшей истории (т-ра в колонне, скорость отбора)
       в виде графика. В отдельном окне.
'''
#-------------------------------------------------------------------------------
